import asyncio
import logging
import os

from bellows.exception import ControllerError, EzspError
import bellows.multicast
import bellows.types as t
import bellows.zigbee.util
from serial import SerialException
import voluptuous as vol
import zigpy.application
import zigpy.device
from zigpy.quirks import CustomDevice, CustomEndpoint
from zigpy.types import BroadcastAddress
import zigpy.util
import zigpy.zdo
import zigpy.zdo.types as zdo_t

APS_ACK_TIMEOUT = 120
CONF_PARAM_SRC_RTG = "source_routing"
CONFIG_SCHEMA = zigpy.application.CONFIG_SCHEMA.extend(
    {vol.Optional(CONF_PARAM_SRC_RTG, default=False): bellows.zigbee.util.cv_boolean}
)
EZSP_DEFAULT_RADIUS = 0
EZSP_MULTICAST_NON_MEMBER_RADIUS = 3
MAX_WATCHDOG_FAILURES = 4
MTOR_MIN_INTERVAL = 600
MTOR_MAX_INTERVAL = 1800
RESET_ATTEMPT_BACKOFF_TIME = 5
WATCHDOG_WAKE_PERIOD = 10

LOGGER = logging.getLogger(__name__)


class ControllerApplication(zigpy.application.ControllerApplication):
    direct = t.EmberOutgoingMessageType.OUTGOING_DIRECT

    def __init__(self, ezsp, database_file=None, config=None):
        if config is None:
            config = {}
        super().__init__(database_file=database_file, config=CONFIG_SCHEMA(config))
        self._ctrl_event = asyncio.Event()
        self._ezsp = ezsp
        self._multicast = bellows.multicast.Multicast(ezsp)
        self._pending = zigpy.util.Requests()
        self._watchdog_task = None
        self._reset_task = None
        self._in_flight_msg = None
        self._tx_options = t.EmberApsOption(
            t.EmberApsOption.APS_OPTION_RETRY
            | t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
        )
        self.use_source_routing = self.config[CONF_PARAM_SRC_RTG]

    @property
    def controller_event(self):
        """Return asyncio.Event for controller app."""
        return self._ctrl_event

    @property
    def is_controller_running(self):
        """Return True if controller was successfully initialized."""
        return self.controller_event.is_set() and self._ezsp.is_ezsp_running

    @property
    def multicast(self):
        """Return EZSP MulticastController."""
        return self._multicast

    async def initialize(self):
        """Perform basic NCP initialization steps"""
        e = self._ezsp

        await e.reset()
        await e.version()

        c = t.EzspConfigId
        await self._cfg(c.CONFIG_STACK_PROFILE, 2)
        await self._cfg(c.CONFIG_SECURITY_LEVEL, 5)
        await self._cfg(c.CONFIG_SUPPORTED_NETWORKS, 1)
        zdo = (
            t.EmberZdoConfigurationFlags.APP_RECEIVES_SUPPORTED_ZDO_REQUESTS
            | t.EmberZdoConfigurationFlags.APP_HANDLES_UNSUPPORTED_ZDO_REQUESTS
        )
        await self._cfg(c.CONFIG_APPLICATION_ZDO_FLAGS, zdo)
        await self._cfg(c.CONFIG_PAN_ID_CONFLICT_REPORT_THRESHOLD, 2)
        await self._cfg(c.CONFIG_TRUST_CENTER_ADDRESS_CACHE_SIZE, 2)
        await self._cfg(c.CONFIG_ADDRESS_TABLE_SIZE, 16)
        await self._cfg(c.CONFIG_SOURCE_ROUTE_TABLE_SIZE, 8)
        await self._cfg(c.CONFIG_MAX_END_DEVICE_CHILDREN, 32)
        await self._cfg(c.CONFIG_INDIRECT_TRANSMISSION_TIMEOUT, 7680)
        await self._cfg(c.CONFIG_KEY_TABLE_SIZE, 4)
        await self._cfg(c.CONFIG_TRANSIENT_KEY_TIMEOUT_S, 180, True)
        if self._ezsp.ezsp_version >= 7:
            await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT, 8)
        else:
            await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT, 60)
            await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT, 8)
        await self._cfg(c.CONFIG_MULTICAST_TABLE_SIZE, self.multicast.TABLE_SIZE)
        await self._cfg(c.CONFIG_PACKET_BUFFER_COUNT, 0xFF)

        status, count = await e.getConfigurationValue(
            c.CONFIG_APS_UNICAST_MESSAGE_COUNT
        )
        assert status == t.EmberStatus.SUCCESS
        self._in_flight_msg = asyncio.Semaphore(count)
        LOGGER.debug("APS_UNICAST_MESSAGE_COUNT is set to %s", count)

        await self.add_endpoint(
            output_clusters=[zigpy.zcl.clusters.security.IasZone.cluster_id]
        )

    async def add_endpoint(
        self,
        endpoint=1,
        profile_id=zigpy.profiles.zha.PROFILE_ID,
        device_id=0xBEEF,
        app_flags=0x00,
        input_clusters=[],
        output_clusters=[],
    ):
        """Add endpoint."""
        res = await self._ezsp.addEndpoint(
            endpoint,
            profile_id,
            device_id,
            app_flags,
            len(input_clusters),
            len(output_clusters),
            input_clusters,
            output_clusters,
        )
        LOGGER.debug("Ezsp adding endpoint: %s", res)

    async def startup(self, auto_form=False):
        """Perform a complete application startup"""
        await self.initialize()
        e = self._ezsp

        await self.set_source_routing()
        v = await e.networkInit()
        if v[0] != t.EmberStatus.SUCCESS:
            if not auto_form:
                raise Exception("Could not initialize network")
            await self.form_network()

        v = await e.getNetworkParameters()
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
        if v[1] != t.EmberNodeType.COORDINATOR:
            if not auto_form:
                raise Exception("Network not configured as coordinator")

            LOGGER.info("Forming network")
            await self._ezsp.leaveNetwork()
            await asyncio.sleep(1)  # TODO
            await self.form_network()

        await self._policy()
        nwk = await e.getNodeId()
        self._nwk = nwk[0]
        ieee = await e.getEui64()
        self._ieee = ieee[0]

        e.add_callback(self.ezsp_callback_handler)
        self.controller_event.set()
        self._watchdog_task = asyncio.ensure_future(self._watchdog())

        self.handle_join(self.nwk, self.ieee, 0)
        LOGGER.debug("EZSP nwk=0x%04x, IEEE=%s", self._nwk, str(self._ieee))

        await self.multicast.startup(self.get_device(self.ieee))

    async def set_source_routing(self) -> None:
        res = await self._ezsp.setConcentrator(
            self.use_source_routing,
            t.EmberConcentratorType.HIGH_RAM_CONCENTRATOR,
            MTOR_MIN_INTERVAL,
            MTOR_MAX_INTERVAL,
            2,
            5,
            0,
        )
        LOGGER.debug("Set concentrator type: %s", res)
        if res[0] != t.EmberStatus.SUCCESS:
            LOGGER.warning(
                "Couldn't set concentrator type %s: %s", self.use_source_routing, res
            )

    async def shutdown(self):
        """Shutdown and cleanup ControllerApplication."""
        LOGGER.info("Shutting down ControllerApplication")
        self.controller_event.clear()
        if self._watchdog_task and not self._watchdog_task.done():
            LOGGER.debug("Cancelling watchdog")
            self._watchdog_task.cancel()
        if self._reset_task and not self._reset_task.done():
            self._reset_task.cancel()
        self._ezsp.close()

    async def form_network(self, channel=15, pan_id=None, extended_pan_id=None):
        channel = t.uint8_t(channel)

        if pan_id is None:
            pan_id = t.uint16_t.from_bytes(os.urandom(2), "little")
        pan_id = t.uint16_t(pan_id)

        if extended_pan_id is None:
            extended_pan_id = t.fixed_list(8, t.uint8_t)([t.uint8_t(0)] * 8)

        initial_security_state = bellows.zigbee.util.zha_security(controller=True)
        v = await self._ezsp.setInitialSecurityState(initial_security_state)
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check

        parameters = t.EmberNetworkParameters()
        parameters.panId = pan_id
        parameters.extendedPanId = extended_pan_id
        parameters.radioTxPower = t.uint8_t(8)
        parameters.radioChannel = channel
        parameters.joinMethod = t.EmberJoinMethod.USE_MAC_ASSOCIATION
        parameters.nwkManagerId = t.EmberNodeId(0)
        parameters.nwkUpdateId = t.uint8_t(0)
        parameters.channels = t.uint32_t(0)

        await self._ezsp.formNetwork(parameters)
        await self._ezsp.setValue(t.EzspValueId.VALUE_STACK_TOKEN_WRITING, 1)

    async def _cfg(self, config_id, value, optional=False):
        v = await self._ezsp.setConfigurationValue(config_id, value)
        if not optional:
            assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check

    async def _policy(self):
        """Set up the policies for what the NCP should do"""
        e = self._ezsp
        v = await e.setPolicy(
            t.EzspPolicyId.TC_KEY_REQUEST_POLICY,
            t.EzspDecisionId.GENERATE_NEW_TC_LINK_KEY,
        )
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
        v = await e.setPolicy(
            t.EzspPolicyId.APP_KEY_REQUEST_POLICY,
            t.EzspDecisionId.ALLOW_APP_KEY_REQUESTS,
        )
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
        v = await e.setPolicy(
            t.EzspPolicyId.TRUST_CENTER_POLICY,
            t.EzspDecisionId.ALLOW_PRECONFIGURED_KEY_JOINS,
        )
        assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check

    async def force_remove(self, dev):
        # This should probably be delivered to the parent device instead
        # of the device itself.
        await self._ezsp.removeDevice(dev.nwk, dev.ieee, dev.ieee)

    def ezsp_callback_handler(self, frame_name, args):
        LOGGER.debug("Received %s frame with %s", frame_name, args)
        if frame_name == "incomingMessageHandler":
            self._handle_frame(*args)
        elif frame_name == "messageSentHandler":
            if args[4] != t.EmberStatus.SUCCESS:
                self._handle_frame_failure(*args)
            else:
                self._handle_frame_sent(*args)
        elif frame_name == "trustCenterJoinHandler":
            if args[2] == t.EmberDeviceUpdate.DEVICE_LEFT:
                self.handle_leave(args[0], args[1])
        elif frame_name == "incomingRouteRecordHandler":
            self.handle_route_record(*args)
        elif frame_name == "incomingRouteErrorHandler":
            self.handle_route_error(*args)
        elif frame_name == "_reset_controller_application":
            self._handle_reset_request(*args)

    def _handle_frame(
        self,
        message_type,
        aps_frame,
        lqi,
        rssi,
        sender,
        binding_index,
        address_index,
        message,
    ):
        if (
            aps_frame.clusterId == zdo_t.ZDOCmd.Device_annce
            and aps_frame.destinationEndpoint == 0
        ):
            nwk, rest = t.uint16_t.deserialize(message[1:])
            ieee, _ = t.EmberEUI64.deserialize(rest)
            LOGGER.info("ZDO Device announce: 0x%04x, %s", nwk, ieee)
            self.handle_join(nwk, ieee, 0)
        try:
            device = self.get_device(nwk=sender)
        except KeyError:
            LOGGER.debug("No such device %s", sender)
            return

        device.radio_details(lqi, rssi)
        self.handle_message(
            device,
            aps_frame.profileId,
            aps_frame.clusterId,
            aps_frame.sourceEndpoint,
            aps_frame.destinationEndpoint,
            message,
        )

    def _handle_frame_failure(
        self, message_type, destination, aps_frame, message_tag, status, message
    ):
        try:
            request = self._pending[message_tag]
            request.result.set_result((status, "message send failure"))
        except KeyError:
            LOGGER.debug(
                "Unexpected message send failure for message tag %s", message_tag
            )
        except asyncio.InvalidStateError as exc:
            LOGGER.debug(
                (
                    "Invalid state on future for message tag %s "
                    "- probably duplicate response: %s"
                ),
                message_tag,
                exc,
            )

    def _handle_frame_sent(
        self, message_type, destination, aps_frame, message_tag, status, message
    ):
        try:
            request = self._pending[message_tag]
            request.result.set_result(
                (t.EmberStatus.SUCCESS, "message sent successfully")
            )
        except KeyError:
            LOGGER.debug("Unexpected message send notification tag: %s", message_tag)
        except asyncio.InvalidStateError as exc:
            LOGGER.debug(
                (
                    "Invalid state on future for message tag %s "
                    "- probably duplicate response: %s"
                ),
                message_tag,
                exc,
            )

    def _handle_reset_request(self, error):
        """Reinitialize application controller."""
        LOGGER.debug("Resetting ControllerApplication. Cause: '%s'", error)
        self.controller_event.clear()
        if self._reset_task:
            LOGGER.debug("Preempting ControllerApplication reset")
            self._reset_task.cancel()

        self._reset_task = asyncio.ensure_future(self._reset_controller_loop())

    async def _reset_controller_loop(self):
        """Keep trying to reset controller until we succeed."""
        self._watchdog_task.cancel()
        while True:
            try:
                await self._reset_controller()
                break
            except (asyncio.TimeoutError, SerialException) as exc:
                LOGGER.debug("ControllerApplication reset unsuccessful: %s", str(exc))
            await asyncio.sleep(RESET_ATTEMPT_BACKOFF_TIME)

        self._reset_task = None
        LOGGER.debug("ControllerApplication successfully reset")

    async def _reset_controller(self):
        """Reset Controller."""
        self._ezsp.close()
        await asyncio.sleep(0.5)
        await self._ezsp.reconnect()
        await self.startup()

    async def mrequest(
        self,
        group_id,
        profile,
        cluster,
        src_ep,
        sequence,
        data,
        *,
        hops=EZSP_DEFAULT_RADIUS,
        non_member_radius=EZSP_MULTICAST_NON_MEMBER_RADIUS
    ):
        """Submit and send data out as a multicast transmission.

        :param group_id: destination multicast address
        :param profile: Zigbee Profile ID to use for outgoing message
        :param cluster: cluster id where the message is being sent
        :param src_ep: source endpoint id
        :param sequence: transaction sequence number of the message
        :param data: Zigbee message payload
        :param hops: the message will be delivered to all nodes within this number of
                     hops of the sender. A value of zero is converted to MAX_HOPS
        :param non_member_radius: the number of hops that the message will be forwarded
                                  by devices that are not members of the group. A value
                                  of 7 or greater is treated as infinite
        :returns: return a tuple of a status and an error_message. Original requestor
                  has more context to provide a more meaningful error message
        """
        if not self.is_controller_running:
            raise ControllerError("ApplicationController is not running")

        aps_frame = t.EmberApsFrame()
        aps_frame.profileId = t.uint16_t(profile)
        aps_frame.clusterId = t.uint16_t(cluster)
        aps_frame.sourceEndpoint = t.uint8_t(src_ep)
        aps_frame.destinationEndpoint = t.uint8_t(src_ep)
        aps_frame.options = t.EmberApsOption(
            t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
        )
        aps_frame.groupId = t.uint16_t(group_id)
        aps_frame.sequence = t.uint8_t(sequence)
        message_tag = self.get_sequence()

        with self._pending.new(message_tag) as req:
            async with self._in_flight_msg:
                res = await self._ezsp.sendMulticast(
                    aps_frame, hops, non_member_radius, message_tag, data
                )
                if res[0] != t.EmberStatus.SUCCESS:
                    return res[0], "EZSP sendMulticast failure: %s" % (res[0],)

                res = await asyncio.wait_for(req.result, APS_ACK_TIMEOUT)
        return res

    async def request(
        self,
        device,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        use_ieee=False,
    ):
        """Submit and send data out as an unicast transmission.

        :param device: destination device
        :param profile: Zigbee Profile ID to use for outgoing message
        :param cluster: cluster id where the message is being sent
        :param src_ep: source endpoint id
        :param dst_ep: destination endpoint id
        :param sequence: transaction sequence number of the message
        :param data: Zigbee message payload
        :param expect_reply: True if this is essentially a request
        :param use_ieee: use EUI64 for destination addressing
        :returns: return a tuple of a status and an error_message. Original requestor
                  has more context to provide a more meaningful error message
        """
        if not self.is_controller_running:
            raise ControllerError("ApplicationController is not running")

        aps_frame = t.EmberApsFrame()
        aps_frame.profileId = t.uint16_t(profile)
        aps_frame.clusterId = t.uint16_t(cluster)
        aps_frame.sourceEndpoint = t.uint8_t(src_ep)
        aps_frame.destinationEndpoint = t.uint8_t(dst_ep)
        aps_frame.options = self._tx_options
        aps_frame.groupId = t.uint16_t(0)
        aps_frame.sequence = t.uint8_t(sequence)
        message_tag = self.get_sequence()

        if use_ieee:
            LOGGER.warning(
                ("EUI64 addressing is not currently supported, " "reverting to NWK")
            )
        with self._pending.new(message_tag) as req:
            async with self._in_flight_msg:
                if expect_reply and device.node_desc.is_end_device in (True, None):
                    LOGGER.debug(
                        "Extending timeout for %s/0x%04x", device.ieee, device.nwk
                    )
                    await self._ezsp.setExtendedTimeout(device.ieee, True)
                if self.use_source_routing and device.relays is not None:
                    res = await self._ezsp.setSourceRoute(device.nwk, device.relays)
                    if res[0] != t.EmberStatus.SUCCESS:
                        LOGGER.warning(
                            "Couldn't set source route for %s: %s", device.nwk, res
                        )
                    else:
                        aps_frame.options = t.EmberApsOption(
                            aps_frame.options
                            ^ t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
                        )
                        LOGGER.debug(
                            "Set source route for %s to %s: %s",
                            device.nwk,
                            device.relays,
                            res,
                        )
                res = await self._ezsp.sendUnicast(
                    self.direct, device.nwk, aps_frame, message_tag, data
                )
                if res[0] != t.EmberStatus.SUCCESS:
                    return res[0], "EZSP sendUnicast failure: %s" % (res[0],)

                res = await asyncio.wait_for(req.result, APS_ACK_TIMEOUT)
        return res

    def permit_ncp(self, time_s=60):
        assert 0 <= time_s <= 254
        return self._ezsp.permitJoining(time_s)

    async def permit_with_key(self, node, code, time_s=60):
        if type(node) is not t.EmberEUI64:
            node = t.EmberEUI64([t.uint8_t(p) for p in node])

        key = zigpy.util.convert_install_code(code)
        if key is None:
            raise Exception("Invalid install code")

        link_key = t.EmberKeyData()
        link_key.contents = key
        v = await self._ezsp.addTransientLinkKey(node, link_key)

        if v[0] != t.EmberStatus.SUCCESS:
            raise Exception("Failed to set link key")

        v = await self._ezsp.setPolicy(
            t.EzspPolicyId.TC_KEY_REQUEST_POLICY,
            t.EzspDecisionId.GENERATE_NEW_TC_LINK_KEY,
        )
        if v[0] != t.EmberStatus.SUCCESS:
            raise Exception(
                "Failed to change policy to allow generation of new trust center keys"
            )

        return await self.permit(time_s)

    async def broadcast(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        grpid,
        radius,
        sequence,
        data,
        broadcast_address=BroadcastAddress.RX_ON_WHEN_IDLE,
    ):
        """Submit and send data out as an unicast transmission.

        :param profile: Zigbee Profile ID to use for outgoing message
        :param cluster: cluster id where the message is being sent
        :param src_ep: source endpoint id
        :param dst_ep: destination endpoint id
        :param: grpid: group id to address the broadcast to
        :param radius: max radius of the broadcast
        :param sequence: transaction sequence number of the message
        :param data: zigbee message payload
        :param timeout: how long to wait for transmission ACK
        :param broadcast_address: broadcast address.
        :returns: return a tuple of a status and an error_message. Original requestor
                  has more context to provide a more meaningful error message
        """
        if not self.is_controller_running:
            raise ControllerError("ApplicationController is not running")

        aps_frame = t.EmberApsFrame()
        aps_frame.profileId = t.uint16_t(profile)
        aps_frame.clusterId = t.uint16_t(cluster)
        aps_frame.sourceEndpoint = t.uint8_t(src_ep)
        aps_frame.destinationEndpoint = t.uint8_t(dst_ep)
        aps_frame.options = t.EmberApsOption(t.EmberApsOption.APS_OPTION_NONE)
        aps_frame.groupId = t.uint16_t(grpid)
        aps_frame.sequence = t.uint8_t(sequence)
        message_tag = self.get_sequence()

        with self._pending.new(message_tag) as req:
            async with self._in_flight_msg:
                res = await self._ezsp.sendBroadcast(
                    broadcast_address, aps_frame, radius, message_tag, data
                )
                if res[0] != t.EmberStatus.SUCCESS:
                    return res[0], "broadcast send failure"

                # Wait for messageSentHandler message
                res = await asyncio.wait_for(req.result, timeout=APS_ACK_TIMEOUT)
        return res

    async def _watchdog(self):
        """Watchdog handler."""
        LOGGER.debug("Starting EZSP watchdog")
        failures = 0
        await asyncio.sleep(WATCHDOG_WAKE_PERIOD)
        while True:
            try:
                await asyncio.wait_for(
                    self.controller_event.wait(), timeout=WATCHDOG_WAKE_PERIOD * 2
                )
                await self._ezsp.nop()
                failures = 0
            except (asyncio.TimeoutError, EzspError) as exc:
                LOGGER.warning("Watchdog heartbeat timeout: %s", str(exc))
                failures += 1
                if failures > MAX_WATCHDOG_FAILURES:
                    break
            await asyncio.sleep(WATCHDOG_WAKE_PERIOD)

        self._handle_reset_request(
            "Watchdog timeout. Heartbeat timeouts: {}".format(failures)
        )

    def handle_route_record(
        self,
        nwk: t.EmberNodeId,
        ieee: t.EmberEUI64,
        lqi: t.uint8_t,
        rssi: t.int8s,
        relays: t.LVList(t.EmberNodeId),
    ) -> None:
        LOGGER.debug(
            "Processing route record request: %s", (nwk, ieee, lqi, rssi, relays)
        )
        try:
            dev = self.get_device(ieee=ieee)
        except KeyError:
            LOGGER.debug("Why we don't have a device for %s ieee and %s NWK", ieee, nwk)
            self.handle_join(nwk, ieee, 0)
            return
        dev.relays = relays

    def handle_route_error(self, status: t.EmberStatus, nwk: t.EmberNodeId) -> None:
        LOGGER.debug("Processing route error: status=%s, nwk=%s", status, nwk)
        try:
            dev = self.get_device(nwk=nwk)
        except KeyError:
            LOGGER.debug("No %s device found", nwk)
            return
        dev.relays = None


class EZSPCoordinator(CustomDevice):
    """Zigpy Device representing Coordinator."""

    class EZSPEndpoint(CustomEndpoint):
        async def add_to_group(self, grp_id: int, name: str = None) -> t.EmberStatus:
            if grp_id in self.member_of:
                return t.EmberStatus.SUCCESS

            app = self.device.application
            status = await app.multicast.subscribe(grp_id)
            if status != t.EmberStatus.SUCCESS:
                self.debug("Couldn't subscribe to 0x%04x group", grp_id)
                return status

            group = app.groups.add_group(grp_id, name)
            group.add_member(self)
            return status

        async def remove_from_group(self, grp_id: int) -> t.EmberStatus:
            if grp_id not in self.member_of:
                return t.EmberStatus.SUCCESS

            app = self.device.application
            status = await app.multicast.unsubscribe(grp_id)
            if status != t.EmberStatus.SUCCESS:
                self.debug("Couldn't unsubscribe 0x%04x group", grp_id)
                return status

            app.groups[grp_id].remove_member(self)
            return status

    signature = {
        "endpoints": {
            1: {
                "profile_id": 0x0104,
                "device_type": 0xBEEF,
                "input_clusters": [],
                "output_clusters": [zigpy.zcl.clusters.security.IasZone.cluster_id],
            }
        }
    }

    replacement = {
        "manufacturer": "Silicon Labs",
        "model": "EZSP",
        "endpoints": {1: (EZSPEndpoint, {})},
    }



import asyncio
import time

import click
import pure_pcapy

from . import util
from .main import main


@main.command()
@click.option(
    "-c", "--channel", type=click.IntRange(11, 26), metavar="CHANNEL", required=True
)
@click.option(
    "-w",
    "outfile",
    type=click.Path(writable=True, dir_okay=False),
    metavar="FILE",
    required=True,
)
@click.pass_context
def dump(ctx, channel, outfile):
    """Capture frames on CHANNEL and write to FILE in tcpdump format"""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_dump(ctx, channel, outfile))
    except KeyboardInterrupt:
        captured = ctx.obj.get("captured", 0)
        start_time = ctx.obj.get("start_time", None)
        if start_time:
            duration = time.time() - start_time
            click.echo("\nCaptured %s frames in %0.2fs" % (captured, duration))
    finally:
        if "ezsp" in ctx.obj:
            loop.run_until_complete(ctx.obj["ezsp"].mfglibEnd())
            ctx.obj["ezsp"].close()


async def _dump(ctx, channel, outfile):
    s = await util.setup(ctx.obj["device"], ctx.obj["baudrate"])
    ctx.obj["ezsp"] = s

    v = await s.mfglibStart(True)
    util.check(v[0], "Unable to start mfglib")

    v = await s.mfglibSetChannel(channel)
    util.check(v[0], "Unable to set channel")

    pcap = pure_pcapy.Dumper(outfile, 128, 195)  # DLT_IEEE_15_4

    click.echo("Capture started")
    ctx.obj["start_time"] = time.time()
    ctx.obj["captured"] = 0

    def cb(frame_name, response):
        if frame_name == "mfglibRxHandler":
            data = response[2]
            ts = time.time()
            ts_sec = int(ts)
            ts_usec = int((ts - ts_sec) * 1000000)
            hdr = pure_pcapy.Pkthdr(ts_sec, ts_usec, len(data), len(data))
            pcap.dump(hdr, data)
            ctx.obj["captured"] += 1

    s.add_callback(cb)

    while True:
        await asyncio.sleep(1)



import os

import click

from . import util

CHANNELS = list(range(11, 27))

channel = click.option(
    "-c", "--channel", type=click.IntRange(11, 26), required=True, default=15
)

channels = click.option(
    "-C",
    "--channels",
    type=util.CSVParamType(11, 26),
    metavar="CHANNELS",
    default=",".join(map(str, CHANNELS)),
    show_default=True,
)

device = click.option(
    "-d", "--device", type=click.STRING, envvar="EZSP_DEVICE", required=True
)

baudrate = click.option(
    "-b", "--baudrate", type=click.INT, envvar="EZSP_BAUDRATE", default=57600
)

database_file = click.option(
    "-D",
    "--database",
    type=click.Path(exists=True, dir_okay=False, writable=True),
    required=True,
    default=os.path.join(click.get_app_dir("bellows"), "app.db"),
)

duration_ms = click.option(
    "-t",
    "--duration",
    "duration_ms",
    type=click.INT,
    metavar="MILLISECONDS",
    default=50,
    show_default=True,
)

duration_s = click.option(
    "-t",
    "--duration",
    "duration_s",
    type=click.INT,
    metavar="SECONDS",
    default=30,
    show_default=True,
)

extended_pan = click.option("-E", "--extended-pan-id", type=click.STRING)

pan = click.option("-P", "--pan-id", type=click.IntRange(0, 65535))

manufacturer = click.option(
    "-m",
    "--manufacturer",
    type=click.IntRange(0, 0xFFFF),
    default=None,
    help="send a manufacturer specific command",
)



import bellows.types as t

COMMANDS = {
    # 4. Configuration frames
    "version": (0x00, (t.uint8_t,), (t.uint8_t, t.uint8_t, t.uint16_t)),
    "getConfigurationValue": (0x52, (t.EzspConfigId,), (t.EzspStatus, t.uint16_t)),
    "setConfigurationValue": (0x53, (t.EzspConfigId, t.uint16_t), (t.EzspStatus,)),
    "addEndpoint": (
        0x02,
        (
            t.uint8_t,
            t.uint16_t,
            t.uint16_t,
            t.uint8_t,
            t.uint8_t,
            t.uint8_t,
            t.List(t.uint16_t),
            t.List(t.uint16_t),
        ),
        (t.EzspStatus,),
    ),
    "setPolicy": (0x55, (t.EzspPolicyId, t.EzspDecisionId), (t.EzspStatus,)),
    "getPolicy": (0x56, (t.EzspPolicyId,), (t.EzspStatus, t.EzspDecisionId)),
    "getValue": (0xAA, (t.EzspValueId,), (t.EzspStatus, t.LVBytes)),
    "getExtendedValue": (
        0x03,
        (t.EzspExtendedValueId, t.uint32_t),
        (t.EzspStatus, t.LVBytes),
    ),
    "setValue": (0xAB, (t.EzspValueId, t.LVBytes), (t.EzspStatus,)),
    "setGpioCurrentConfiguration": (
        0xAC,
        (t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EzspStatus,),
    ),
    "setGpioPowerUpDownConfiguration": (
        0xAD,
        (t.uint8_t, t.uint8_t, t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EzspStatus,),
    ),
    "setGpioRadioPowerMask": (0xAE, (t.uint32_t,), ()),
    "setCtune": (0xF5, (t.uint16_t,), ()),
    "getCtune": (0xF6, (), (t.uint16_t,)),
    "setChannelMap": (0xF7, (t.uint8_t, t.uint8_t), ()),
    # 5. Utilities Frames
    "nop": (0x05, (), ()),
    "echo": (0x81, (t.LVBytes,), (t.LVBytes,)),
    "invalidCommand": (0x58, (), (t.EzspStatus,)),
    "callback": (0x06, (), ()),
    "noCallbacks": (0x07, (), ()),
    "setToken": (0x09, (t.uint8_t, t.fixed_list(8, t.uint8_t)), (t.EmberStatus,)),
    "getToken": (0x0A, (t.uint8_t,), (t.EmberStatus, t.fixed_list(8, t.uint8_t))),
    "getMfgToken": (0x0B, (t.EzspMfgTokenId,), (t.LVBytes,)),
    "setMfgToken": (0x0C, (t.EzspMfgTokenId, t.LVBytes), (t.EmberStatus,)),
    "stackTokenChangedHandler": (0x0D, (), (t.uint16_t,)),
    "getRandomNumber": (0x49, (), (t.EmberStatus, t.uint16_t)),
    "setTimer": (
        0x0E,
        (t.uint8_t, t.uint16_t, t.EmberEventUnits, t.Bool),
        (t.EmberStatus,),
    ),
    "getTimer": (0x4E, (t.uint8_t,), (t.uint16_t, t.EmberEventUnits, t.Bool)),
    "timerHandler": (0x0F, (), (t.uint8_t,)),
    "debugWrite": (0x12, (t.Bool, t.LVBytes), (t.EmberStatus,)),
    "readAndClearCounters": (
        0x65,
        (),
        (t.fixed_list(t.EmberCounterType.COUNTER_TYPE_COUNT, t.uint16_t),),
    ),
    "readCounters": (
        0xF1,
        (),
        (t.fixed_list(t.EmberCounterType.COUNTER_TYPE_COUNT, t.uint16_t),),
    ),
    "counterRolloverHandler": (0xF2, (), (t.EmberCounterType,)),
    "delayTest": (0x9D, (t.uint16_t,), ()),
    "getLibraryStatus": (0x01, (t.uint8_t,), (t.EmberLibraryStatus,)),
    "getXncpInfo": (0x13, (), (t.EmberStatus, t.uint16_t, t.uint16_t)),
    "customFrame": (0x47, (t.LVBytes,), (t.EmberStatus, t.LVBytes)),
    "customFrameHandler": (0x54, (), (t.LVBytes,)),
    "getEui64": (0x26, (), (t.EmberEUI64,)),
    "getNodeId": (0x27, (), (t.EmberNodeId,)),
    "networkInit": (0x17, (), (t.EmberStatus,)),
    # 6. Networking Frames
    "setManufacturerCode": (0x15, (t.uint16_t,), ()),
    "setPowerDescriptor": (0x16, (t.uint16_t,), ()),
    "networkInitExtended": (0x70, (t.EmberNetworkInitStruct,), (t.EmberStatus,)),
    "networkState": (0x18, (), (t.EmberNetworkStatus,)),
    "stackStatusHandler": (0x19, (), (t.EmberStatus,)),
    "startScan": (
        0x1A,
        (t.EzspNetworkScanType, t.uint32_t, t.uint8_t),
        (t.EmberStatus,),
    ),
    "energyScanResultHandler": (0x48, (), (t.uint8_t, t.int8s)),
    "networkFoundHandler": (0x1B, (), (t.EmberZigbeeNetwork, t.uint8_t, t.int8s)),
    "scanCompleteHandler": (0x1C, (), (t.uint8_t, t.EmberStatus)),
    "stopScan": (0x1D, (), (t.EmberStatus,)),
    "formNetwork": (0x1E, (t.EmberNetworkParameters,), (t.EmberStatus,)),
    "joinNetwork": (
        0x1F,
        (t.EmberNodeType, t.EmberNetworkParameters),
        (t.EmberStatus,),
    ),
    "leaveNetwork": (0x20, (), (t.EmberStatus,)),
    "findAndRejoinNetwork": (0x21, (t.Bool, t.uint32_t), (t.EmberStatus,)),
    "permitJoining": (0x22, (t.uint8_t,), (t.EmberStatus,)),
    "childJoinHandler": (
        0x23,
        (),
        (t.uint8_t, t.Bool, t.EmberNodeId, t.EmberEUI64, t.EmberNodeType),
    ),
    "energyScanRequest": (
        0x9C,
        (t.EmberNodeId, t.uint32_t, t.uint8_t, t.uint16_t),
        (t.EmberStatus,),
    ),
    "getNetworkParameters": (
        0x28,
        (),
        (t.EmberStatus, t.EmberNodeType, t.EmberNetworkParameters),
    ),
    "getParentChildParameters": (0x29, (), (t.uint8_t, t.EmberEUI64, t.EmberNodeId)),
    "getChildData": (
        0x4A,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberNodeId, t.EmberEUI64, t.EmberNodeType),
    ),
    "getNeighbor": (0x79, (t.uint8_t,), (t.EmberStatus, t.EmberNeighborTableEntry)),
    "neighborCount": (0x7A, (), (t.uint8_t,)),
    "getRouteTableEntry": (0x7B, (t.uint8_t,), (t.EmberStatus, t.EmberRouteTableEntry)),
    "setRadioPower": (0x99, (t.int8s,), (t.EmberStatus,)),
    "setRadioChannel": (0x9A, (t.uint8_t,), (t.EmberStatus,)),
    "setConcentrator": (
        0x10,
        (t.Bool, t.uint16_t, t.uint16_t, t.uint16_t, t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EmberStatus,),
    ),
    # 7. Binding Frames
    "clearBindingTable": (0x2A, (), (t.EmberStatus,)),
    "setBinding": (0x2B, (t.uint8_t, t.EmberBindingTableEntry), (t.EmberStatus,)),
    "getBinding": (0x2C, (t.uint8_t,), (t.EmberStatus, t.EmberBindingTableEntry)),
    "deleteBinding": (0x2D, (t.uint8_t,), (t.EmberStatus,)),
    "bindingIsActive": (0x2E, (t.uint8_t,), (t.Bool,)),
    "getBindingRemoteNodeId": (0x2F, (t.uint8_t,), (t.EmberNodeId,)),
    "setBindingRemoteNodeId": (0x30, (t.uint8_t,), ()),
    "remoteSetBindingHandler": (0x31, (), (t.EmberBindingTableEntry,)),
    "remoteDeleteBindingHandler": (0x32, (), (t.uint8_t, t.EmberStatus)),
    # 8. Messaging Frames
    "maximumPayloadLength": (0x33, (), (t.uint8_t,)),
    "sendUnicast": (
        0x34,
        (
            t.EmberOutgoingMessageType,
            t.EmberNodeId,
            t.EmberApsFrame,
            t.uint8_t,
            t.LVBytes,
        ),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendBroadcast": (
        0x36,
        (t.EmberNodeId, t.EmberApsFrame, t.uint8_t, t.uint8_t, t.LVBytes),
        (t.EmberStatus, t.uint8_t),
    ),
    "proxyBroadcast": (
        0x37,
        (
            t.EmberNodeId,
            t.EmberNodeId,
            t.uint8_t,
            t.EmberApsFrame,
            t.uint8_t,
            t.uint8_t,
            t.LVBytes,
        ),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendMulticast": (
        0x38,
        (t.EmberApsFrame, t.uint8_t, t.uint8_t, t.uint8_t, t.LVBytes),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendReply": (0x39, (t.EmberNodeId, t.EmberApsFrame, t.LVBytes), (t.EmberStatus,)),
    "messageSentHandler": (
        0x3F,
        (),
        (
            t.EmberOutgoingMessageType,
            t.uint16_t,
            t.EmberApsFrame,
            t.uint8_t,
            t.EmberStatus,
            t.LVBytes,
        ),
    ),
    "sendManyToOneRouteRequest": (0x41, (t.uint16_t, t.uint8_t), (t.EmberStatus,)),
    "pollForData": (0x42, (t.uint16_t, t.EmberEventUnits, t.uint8_t), (t.EmberStatus,)),
    "pollCompleteHandler": (0x43, (), (t.EmberStatus,)),
    "pollHandler": (0x44, (), (t.EmberNodeId,)),
    "incomingSenderEui64Handler": (0x62, (), (t.EmberEUI64,)),
    "incomingMessageHandler": (
        0x45,
        (),
        (
            t.EmberIncomingMessageType,
            t.EmberApsFrame,
            t.uint8_t,
            t.int8s,
            t.EmberNodeId,
            t.uint8_t,
            t.uint8_t,
            t.LVBytes,
        ),
    ),
    "incomingRouteRecordHandler": (
        0x59,
        (),
        (t.EmberNodeId, t.EmberEUI64, t.uint8_t, t.int8s, t.LVList(t.EmberNodeId)),
    ),
    "changeSourceRouteHandler": (0xC4, (), (t.EmberNodeId, t.EmberNodeId, t.Bool)),
    "setSourceRoute": (
        0x5A,
        (t.EmberNodeId, t.LVList(t.EmberNodeId)),
        (t.EmberStatus,),
    ),
    "incomingManyToOneRouteRequestHandler": (
        0x7D,
        (),
        (t.EmberNodeId, t.EmberEUI64, t.uint8_t),
    ),
    "incomingRouteErrorHandler": (0x80, (), (t.EmberStatus, t.EmberNodeId)),
    "addressTableEntryIsActive": (0x5B, (t.uint8_t,), (t.Bool,)),
    "setAddressTableRemoteEui64": (0x5C, (t.uint8_t, t.EmberEUI64), (t.EmberStatus,)),
    "setAddressTableRemoteNodeId": (0x5D, (t.uint8_t, t.EmberNodeId), ()),
    "getAddressTableRemoteEui64": (0x5E, (t.uint8_t,), (t.EmberEUI64,)),
    "getAddressTableRemoteNodeId": (0x5F, (t.uint8_t,), (t.EmberNodeId,)),
    "setExtendedTimeout": (0x7E, (t.EmberEUI64, t.Bool), ()),
    "getExtendedTimeout": (0x7F, (t.EmberEUI64,), (t.Bool,)),
    "replaceAddressTableEntry": (
        0x82,
        (t.uint8_t, t.EmberEUI64, t.EmberNodeId, t.Bool),
        (t.EmberStatus, t.EmberEUI64, t.EmberNodeId, t.Bool),
    ),
    "lookupNodeIdByEui64": (0x60, (t.EmberEUI64,), (t.EmberNodeId,)),
    "lookupEui64ByNodeId": (0x61, (t.EmberNodeId,), (t.EmberStatus, t.EmberEUI64)),
    "getMulticastTableEntry": (
        0x63,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberMulticastTableEntry),
    ),
    "setMulticastTableEntry": (
        0x64,
        (t.uint8_t, t.EmberMulticastTableEntry),
        (t.EmberStatus,),
    ),
    "idConflictHandler": (0x7C, (), (t.EmberNodeId,)),
    "sendRawMessage": (0x96, (t.LVBytes,), (t.EmberStatus,)),
    "macPassthroughMessageHandler": (
        0x97,
        (),
        (t.EmberMacPassthroughType, t.uint8_t, t.int8s, t.LVBytes),
    ),
    "macFilterMatchMessageHandler": (
        0x46,
        (),
        (t.uint8_t, t.EmberMacPassthroughType, t.uint8_t, t.int8s, t.LVBytes),
    ),
    "rawTransmitCompleteHandler": (0x98, (), (t.EmberStatus,)),
    # 9. Security Frames
    "setInitialSecurityState": (0x68, (t.EmberInitialSecurityState,), (t.EmberStatus,)),
    "getCurrentSecurityState": (0x69, (), (t.EmberStatus, t.EmberCurrentSecurityState)),
    "getKey": (0x6A, (t.EmberKeyType,), (t.EmberStatus, t.EmberKeyStruct)),
    "switchNetworkKeyHandler": (0x6E, (), (t.uint8_t,)),
    "getKeyTableEntry": (0x71, (t.uint8_t,), (t.EmberStatus, t.EmberKeyStruct)),
    "setKeyTableEntry": (
        0x72,
        (t.uint8_t, t.EmberEUI64, t.Bool, t.EmberKeyData),
        (t.EmberStatus,),
    ),
    "findKeyTableEntry": (0x75, (t.EmberEUI64, t.Bool), (t.uint8_t,)),
    "addOrUpdateKeyTableEntry": (
        0x66,
        (t.EmberEUI64, t.Bool, t.EmberKeyData),
        (t.EmberStatus,),
    ),
    "eraseKeyTableEntry": (0x76, (t.uint8_t,), (t.EmberStatus,)),
    "clearKeyTable": (0xB1, (), (t.EmberStatus,)),
    "requestLinkKey": (0x14, (t.EmberEUI64,), (t.EmberStatus,)),
    "zigbeeKeyEstablishmentHandler": (0x9B, (), (t.EmberEUI64, t.EmberKeyStatus)),
    "addTransientLinkKey": (0xAF, (t.EmberEUI64, t.EmberKeyData), (t.EmberStatus,)),
    "clearTransientLinkKeys": (0x6B, (), ()),
    "getTransientLinkKey": (
        0xCE,
        (t.EmberEUI64,),
        (t.EmberStatus, t.EmberTransientKeyData),
    ),
    "setSecurityKey": (
        0xCA,
        (t.EmberKeyData, t.SecureEzspSecurityType),
        (t.EzspStatus,),
    ),
    "setSecurityParameters": (
        0xCB,
        (t.SecureEzspSecurityLevel, t.SecureEzspRandomNumber),
        (t.EzspStatus, t.SecureEzspRandomNumber),
    ),
    "resetToFactoryDefaults": (0xCC, (), (t.EzspStatus,)),
    "getSecurityKeyStatus": (0xCD, (), (t.EzspStatus, t.SecureEzspSecurityType)),
    # 10. Trust Center Frames
    "trustCenterJoinHandler": (
        0x24,
        (),
        (
            t.EmberNodeId,
            t.EmberEUI64,
            t.EmberDeviceUpdate,
            t.EmberJoinDecision,
            t.EmberNodeId,
        ),
    ),
    "broadcastNextNetworkKey": (0x73, (t.EmberKeyData,), (t.EmberStatus,)),
    "broadcastNetworkKeySwitch": (0x74, (), (t.EmberStatus,)),
    "becomeTrustCenter": (0x77, (t.EmberKeyData,), (t.EmberStatus,)),
    "aesMmoHash": (
        0x6F,
        (t.EmberAesMmoHashContext, t.Bool, t.LVBytes),
        (t.EmberStatus, t.EmberAesMmoHashContext),
    ),
    "removeDevice": (
        0xA8,
        (t.EmberNodeId, t.EmberEUI64, t.EmberEUI64),
        (t.EmberStatus,),
    ),
    "unicastNwkKeyUpdate": (
        0xA9,
        (t.EmberNodeId, t.EmberEUI64, t.EmberKeyData),
        (t.EmberStatus,),
    ),
    # 11. Certificate Based Key Exchange (CBKE) Frames
    "generateCbkeKeys": (0xA4, (), (t.EmberStatus,)),
    "generateCbkeKeysHandler": (0x9E, (), (t.EmberStatus, t.EmberPublicKeyData)),
    "calculateSmacs": (
        0x9F,
        (t.Bool, t.EmberCertificateData, t.EmberPublicKeyData),
        (t.EmberStatus,),
    ),
    "calculateSmacsHandler": (
        0xA0,
        (),
        (t.EmberStatus, t.EmberSmacData, t.EmberSmacData),
    ),
    "generateCbkeKeys283k1": (0xE8, (), (t.EmberStatus,)),
    "generateCbkeKeysHandler283k1": (
        0xE9,
        (),
        (t.EmberStatus, t.EmberPublicKey283k1Data),
    ),
    "calculateSmacs283k1": (
        0xEA,
        (t.Bool, t.EmberCertificate283k1Data, t.EmberPublicKey283k1Data),
        (t.EmberStatus,),
    ),
    "calculateSmacsHandler283k1": (
        0xEB,
        (),
        (t.EmberStatus, t.EmberSmacData, t.EmberSmacData),
    ),
    "clearTemporaryDataMaybeStoreLinkKey": (0xA1, (t.Bool,), (t.EmberStatus,)),
    "clearTemporaryDataMaybeStoreLinkKey283k1": (0xEE, (t.Bool,), (t.EmberStatus,)),
    "getCertificate": (0xA5, (), (t.EmberStatus, t.EmberCertificateData)),
    "getCertificate283k1": (0xEC, (), (t.EmberStatus, t.EmberCertificate283k1Data)),
    "dsaSign": (0xA6, (t.LVBytes,), (t.EmberStatus,)),  # Deprecated
    "dsaSignHandler": (0xA7, (), (t.EmberStatus, t.LVBytes)),  # Deprecated
    "dsaVerify": (
        0xA3,
        (t.EmberMessageDigest, t.EmberCertificateData, t.EmberSignatureData),
        (t.EmberStatus,),
    ),
    "dsaVerifyHandler": (0x78, (), (t.EmberStatus,)),
    "dsaVerify283k1": (
        0xB0,
        (t.EmberMessageDigest, t.EmberCertificate283k1Data, t.EmberSignature283k1Data),
        (t.EmberStatus,),
    ),
    "setPreinstalledCbkeData": (
        0xA2,
        (t.EmberPublicKeyData, t.EmberCertificateData, t.EmberPrivateKeyData),
        (t.EmberStatus,),
    ),
    "setPreinstalledCbkeData283k1": (
        0xED,
        (
            t.EmberPublicKey283k1Data,
            t.EmberCertificate283k1Data,
            t.EmberPrivateKey283k1Data,
        ),
        (t.EmberStatus,),
    ),
    # 12. Mfglib Frames
    "mfglibStart": (0x83, (t.Bool,), (t.EmberStatus,)),
    "mfglibEnd": (0x84, (), (t.EmberStatus,)),
    "mfglibStartTone": (0x85, (), (t.EmberStatus,)),
    "mfglibStopTone": (0x86, (), (t.EmberStatus,)),
    "mfglibStartStream": (0x87, (), (t.EmberStatus,)),
    "mfglibStopStream": (0x88, (), (t.EmberStatus,)),
    "mfglibSendPacket": (0x89, (t.LVBytes,), (t.EmberStatus,)),
    "mfglibSetChannel": (0x8A, (t.uint8_t,), (t.EmberStatus,)),
    "mfglibGetChannel": (0x8B, (), (t.uint8_t,)),
    "mfglibSetPower": (0x8C, (t.uint16_t, t.int8s), (t.EmberStatus,)),
    "mfglibGetPower": (0x8D, (), (t.int8s,)),
    "mfglibRxHandler": (0x8E, (), (t.uint8_t, t.int8s, t.LVBytes)),
    # 13. Bootloader Frames
    "launchStandaloneBootloader": (0x8F, (t.uint8_t,), (t.EmberStatus,)),
    "sendBootloadMessage": (0x90, (t.Bool, t.EmberEUI64, t.LVBytes), (t.EmberStatus,)),
    "getStandaloneBootloaderVersionPlatMicroPhy": (
        0x91,
        (),
        (t.uint16_t, t.uint8_t, t.uint8_t, t.uint8_t),
    ),
    "incomingBootloadMessageHandler": (
        0x92,
        (),
        (t.EmberEUI64, t.uint8_t, t.int8s, t.LVBytes),
    ),
    "bootloadTransmitCompleteHandler": (0x93, (), (t.EmberStatus, t.LVBytes)),
    "aesEncrypt": (
        0x94,
        (t.fixed_list(16, t.uint8_t), t.fixed_list(16, t.uint8_t)),
        (t.fixed_list(16, t.uint8_t),),
    ),
    "overrideCurrentChannel": (0x95, (t.uint8_t,), (t.EmberStatus,)),
    # 14. ZLL Frames
    "zllNetworkOps": (
        0xB2,
        (t.EmberZllNetwork, t.EzspZllNetworkOperation, t.int8s),
        (t.EmberStatus,),
    ),
    "zllSetInitialSecurityState": (
        0xB3,
        (t.EmberKeyData, t.EmberZllInitialSecurityState),
        (t.EmberStatus,),
    ),
    "zllStartScan": (0xB4, (t.uint32_t, t.int8s, t.EmberNodeType), (t.EmberStatus,)),
    "zllSetRxOnWhenIdle": (0xB5, (t.uint16_t,), (t.EmberStatus,)),
    "zllNetworkFoundHandler": (
        0xB6,
        (),
        (t.EmberZllNetwork, t.Bool, t.EmberZllDeviceInfoRecord, t.uint8_t, t.int8s),
    ),
    "zllScanCompleteHandler": (0xB7, (), (t.EmberStatus,)),
    "zllAddressAssignmentHandler": (
        0xB8,
        (),
        (t.EmberZllAddressAssignment, t.uint8_t, t.int8s),
    ),
    "setLogicalAndRadioChannel": (0xB9, (t.uint8_t,), (t.EmberStatus,)),
    "getLogicalChannel": (0xBA, (), (t.uint8_t,)),
    "zllTouchLinkTargetHandler": (0xBB, (), (t.EmberZllNetwork,)),
    "zllGetTokens": (
        0xBC,
        (),
        (t.EmberTokTypeStackZllData, t.EmberTokTypeStackZllSecurity),
    ),
    "zllSetDataToken": (0xBD, (t.EmberTokTypeStackZllData,), ()),
    "zllSetNonZllNetwork": (0xBF, (), ()),
    "isZllNetwork": (0xBE, (), (t.Bool,)),
    # 15 RF4CE Frames
    "rf4ceSetPairingTableEntry": (
        0xD0,
        (t.uint8_t, t.EmberRf4cePairingTableEntry),
        (t.EmberStatus,),
    ),
    "rf4ceGetPairingTableEntry": (
        0xD1,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberRf4cePairingTableEntry),
    ),
    "rf4ceDeletePairingTableEntry": (0xD2, (t.uint8_t,), (t.EmberStatus,)),
    "rf4ceKeyUpdate": (0xD3, (t.uint8_t, t.EmberKeyData), (t.EmberStatus,)),
    "rf4ceSend": (
        0xD4,
        (t.uint8_t, t.uint8_t, t.uint16_t, t.EmberRf4ceTxOption, t.uint8_t, t.LVBytes),
        (t.EmberStatus,),
    ),
    "rf4ceIncomingMessageHandler": (
        0xD5,
        (),
        (t.uint8_t, t.uint8_t, t.uint16_t, t.EmberRf4ceTxOption, t.LVBytes),
    ),
    "rf4ceMessageSentHandler": (
        0xD6,
        (),
        (
            t.EmberStatus,
            t.uint8_t,
            t.EmberRf4ceTxOption,
            t.uint8_t,
            t.uint16_t,
            t.uint8_t,
            t.LVBytes,
        ),
    ),
    "rf4ceStart": (
        0xD7,
        (t.EmberRf4ceNodeCapabilities, t.EmberRf4ceVendorInfo, t.int8s),
        (t.EmberStatus,),
    ),
    "rf4ceStop": (0xD8, (), (t.EmberStatus,)),
    "rf4ceDiscovery": (
        0xD9,
        (t.EmberPanId, t.EmberNodeId, t.uint8_t, t.uint16_t, t.LVBytes),
        (t.EmberStatus,),
    ),
    "rf4ceDiscoveryCompleteHandler": (0xDA, (), (t.EmberStatus,)),
    "rf4ceDiscoveryRequestHandler": (
        0xDB,
        (),
        (
            t.EmberEUI64,
            t.uint8_t,
            t.EmberRf4ceVendorInfo,
            t.EmberRf4ceApplicationInfo,
            t.uint8_t,
            t.uint8_t,
        ),
    ),
    "rf4ceDiscoveryResponseHandler": (
        0xDC,
        (),
        (
            t.Bool,
            t.uint8_t,
            t.EmberPanId,
            t.EmberEUI64,
            t.uint8_t,
            t.EmberRf4ceVendorInfo,
            t.EmberRf4ceApplicationInfo,
            t.uint8_t,
            t.uint8_t,
        ),
    ),
    "rf4ceEnableAutoDiscoveryResponse": (0xDD, (t.uint16_t,), (t.EmberStatus,)),
    "rf4ceAutoDiscoveryResponseCompleteHandler": (
        0xDE,
        (),
        (
            t.EmberStatus,
            t.EmberEUI64,
            t.uint8_t,
            t.EmberRf4ceVendorInfo,
            t.EmberRf4ceApplicationInfo,
            t.uint8_t,
        ),
    ),
    "rf4cePair": (
        0xDF,
        (t.uint8_t, t.EmberPanId, t.EmberEUI64, t.uint8_t),
        (t.EmberStatus,),
    ),
    "rf4cePairCompleteHandler": (
        0xE0,
        (),
        (t.EmberStatus, t.uint8_t, t.EmberRf4ceVendorInfo, t.EmberRf4ceApplicationInfo),
    ),
    "rf4cePairRequestHandler": (
        0xE1,
        (),
        (
            t.EmberStatus,
            t.uint8_t,
            t.EmberEUI64,
            t.uint8_t,
            t.EmberRf4ceVendorInfo,
            t.EmberRf4ceApplicationInfo,
            t.uint8_t,
        ),
    ),
    "rf4ceUnpair": (0xE2, (t.uint8_t,), (t.EmberStatus,)),
    "rf4ceUnpairHandler": (0xE3, (), (t.uint8_t,)),
    "rf4ceUnpairCompleteHandler": (0xE4, (), (t.uint8_t,)),
    "rf4ceSetPowerSavingParameters": (0xE5, (t.uint32_t, t.uint32_t), (t.EmberStatus,)),
    "rf4ceSetFrequencyAgilityParameters": (
        0xE6,
        (t.uint8_t, t.uint8_t, t.int8s, t.uint16_t, t.uint8_t),
        (t.EmberStatus,),
    ),
    "rf4ceSetApplicationInfo": (0xE7, (t.EmberRf4ceApplicationInfo,), (t.EmberStatus,)),
    "rf4ceGetApplicationInfo": (0xEF, (), (t.EmberStatus, t.EmberRf4ceApplicationInfo)),
    "rf4ceGetMaxPayload": (0xF3, (t.uint8_t, t.EmberRf4ceTxOption), (t.uint8_t,)),
    "rf4ceGetNetworkParameters": (
        0xF4,
        (),
        (t.EmberStatus, t.EmberNodeType, t.EmberNetworkParameters),
    ),
    # 16 Green Power Frames
    "gpProxyTableProcessGpPairing": (
        0xC9,
        (
            t.uint32_t,
            t.EmberGpAddress,
            t.uint8_t,
            t.uint16_t,
            t.uint16_t,
            t.uint16_t,
            t.fixed_list(8, t.uint8_t),
            t.EmberKeyData,
        ),
        (),
    ),
    "dGpSend": (
        0xC6,
        (t.Bool, t.Bool, t.EmberGpAddress, t.uint8_t, t.LVBytes, t.uint8_t, t.uint16_t),
        (t.EmberStatus,),
    ),
    "dGpSentHandler": (0xC7, (), (t.EmberStatus, t.uint8_t)),
    "gpepIncomingMessageHandler": (
        0xC5,
        (),
        (
            t.EmberStatus,
            t.uint8_t,
            t.uint8_t,
            t.EmberGpAddress,
            t.EmberGpSecurityLevel,
            t.EmberGpKeyType,
            t.Bool,
            t.Bool,
            t.uint32_t,
            t.uint8_t,
            t.uint32_t,
            t.EmberGpSinkListEntry,
            t.LVBytes,
        ),
    ),
}



import asyncio
from concurrent.futures import ThreadPoolExecutor
import functools
import logging
import sys

LOGGER = logging.getLogger(__name__)


class EventLoopThread:
    """ Run a parallel event loop in a separate thread """

    def __init__(self):
        self.loop = None
        self.thread_complete = None

    def run_coroutine_threadsafe(self, coroutine):
        current_loop = asyncio.get_event_loop()
        future = asyncio.run_coroutine_threadsafe(coroutine, self.loop)
        return asyncio.wrap_future(future, loop=current_loop)

    def _thread_main(self, init_task):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        try:
            self.loop.run_until_complete(init_task)
            self.loop.run_forever()
        finally:
            self.loop.close()
            self.loop = None

    async def start(self):
        current_loop = asyncio.get_event_loop()
        if self.loop is not None and not self.loop.is_closed():
            return

        executor_opts = {"max_workers": 1}
        if sys.version_info[:2] >= (3, 6):
            executor_opts["thread_name_prefix"] = __name__
        executor = ThreadPoolExecutor(**executor_opts)

        thread_started_future = current_loop.create_future()

        async def init_task():
            current_loop.call_soon_threadsafe(thread_started_future.set_result, None)

        # Use current loop so current loop has a reference to the long-running thread
        # as one of its tasks
        thread_complete = current_loop.run_in_executor(
            executor, self._thread_main, init_task()
        )
        self.thread_complete = thread_complete
        current_loop.call_soon(executor.shutdown, False)
        await thread_started_future
        return thread_complete

    def force_stop(self):
        if self.loop is not None:
            self.loop.call_soon_threadsafe(self.loop.stop)


class ThreadsafeProxy:
    """ Proxy class which enforces threadsafe non-blocking calls
        This class can be used to wrap an object to ensure any calls
        using that object's methods are done on a particular event loop
    """

    def __init__(self, obj, obj_loop):
        self._obj = obj
        self._obj_loop = obj_loop

    def __getattr__(self, name):
        func = getattr(self._obj, name)
        if not callable(func):
            raise TypeError(
                "Can only use ThreadsafeProxy with callable attributes: {}.{}".format(
                    self._obj.__class__.__name__, name
                )
            )

        def func_wrapper(*args, **kwargs):
            loop = self._obj_loop
            curr_loop = asyncio.get_event_loop()
            call = functools.partial(func, *args, **kwargs)
            if loop == curr_loop:
                return call()
            if loop.is_closed():
                # Disconnected
                LOGGER.warning("Attempted to use a closed event loop")
                return
            if asyncio.iscoroutinefunction(func):
                future = asyncio.run_coroutine_threadsafe(call(), loop)
                return asyncio.wrap_future(future, loop=curr_loop)
            else:

                def check_result_wrapper():
                    result = call()
                    if result is not None:
                        raise TypeError(
                            (
                                "ThreadsafeProxy can only wrap functions with no return"
                                "value \nUse an async method to return values: {}.{}"
                            ).format(self._obj.__class__.__name__, name)
                        )

                loop.call_soon_threadsafe(check_result_wrapper)

        return func_wrapper



"""Setup module for bellows"""

import bellows
from setuptools import find_packages, setup

setup(
    name="bellows-homeassistant",
    version=bellows.__version__,
    description="Library implementing EZSP",
    url="http://github.com/zigpy/bellows",
    author="Russell Cloran",
    author_email="rcloran@gmail.com",
    license="GPL-3.0",
    packages=find_packages(exclude=["*.tests"]),
    entry_points={"console_scripts": ["bellows=bellows.cli.main:main"]},
    install_requires=[
        "Click",
        "click-log==0.2.0",
        "pure_pcapy3==1.0.1",
        "pyserial-asyncio",
        "voluptuous",
        "zigpy-homeassistant>=0.17.0",
    ],
    dependency_links=["https://codeload.github.com/rcloran/pure-pcapy-3/zip/master"],
    tests_require=["asynctest", "pytest", "pytest-asyncio"],
)



import os
from typing import Any

import bellows.types as t
import voluptuous as vol


def zha_security(controller=False):
    empty_key_data = t.EmberKeyData()
    empty_key_data.contents = t.fixed_list(16, t.uint8_t)([t.uint8_t(0)] * 16)
    zha_key = t.EmberKeyData()
    zha_key.contents = t.fixed_list(16, t.uint8_t)(
        [t.uint8_t(c) for c in b"ZigBeeAlliance09"]
    )

    isc = t.EmberInitialSecurityState()
    isc.bitmask = t.uint16_t(
        t.EmberInitialSecurityBitmask.HAVE_PRECONFIGURED_KEY
        | t.EmberInitialSecurityBitmask.REQUIRE_ENCRYPTED_KEY
    )
    isc.preconfiguredKey = zha_key
    isc.networkKey = empty_key_data
    isc.networkKeySequenceNumber = t.uint8_t(0)
    isc.preconfiguredTrustCenterEui64 = t.EmberEUI64([t.uint8_t(0)] * 8)

    if controller:
        isc.bitmask |= (
            t.EmberInitialSecurityBitmask.TRUST_CENTER_GLOBAL_LINK_KEY
            | t.EmberInitialSecurityBitmask.HAVE_NETWORK_KEY
        )
        isc.bitmask = t.uint16_t(isc.bitmask)
        random_key = t.fixed_list(16, t.uint8_t)([t.uint8_t(x) for x in os.urandom(16)])
        isc.networkKey = random_key
    return isc


def cv_boolean(value: Any) -> bool:
    """Validate and coerce a boolean value."""
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        value = value.lower().strip()
        if value in ("1", "true", "yes", "on", "enable"):
            return True
        if value in ("0", "false", "no", "off", "disable"):
            return False
    elif isinstance(value, int):
        return bool(value)
    raise vol.Invalid("invalid boolean value {}".format(value))
